@@ -1,13 +1,13 @@
 from __future__ import print_function
-from os import listdir
-from os.path import isfile, join, dirname, splitext, isdir, basename
+import os
 from struct import unpack
 import zipfile
 import numpy as np
 import csv
 from operator import itemgetter
-from datetime import datetime, timedelta
+from datetime import datetime, timedelta, timezone
 import warnings
+from collections import namedtuple
 
 
 # Error Management
@@ -55,6 +55,9 @@ class CalibrationFileEmptyError(CalibrationFileError):
     pass
 
 
+FrameContainer = namedtuple('FrameContainer', ['header', 'data', 'timestamp', 'valid'])
+
+
 def sat_dtype_to_np_dtype(sat_data_type, sat_field_length):
     if sat_data_type == 'AS':
         # ASCII string (text)
@@ -106,7 +109,7 @@ class Parser:
     Follow Satlantic's Data Format Standard SAT-DN-00134, ver 6.1
     """
 
-    CORE_VARIABLE_TYPES = ['LT', 'LI', 'LU', 'LD', 'LS', 'ED', 'ES', 'EU', 'EV', 'EF']
+    CORE_VARIABLE_TYPES = ['LT', 'LI', 'LU', 'LD', 'LS', 'ED', 'ES', 'EU', 'EV', 'EF', 'UV']
 
     def __init__(self, cal_filename=None, immersed=False):
         # Metadata
@@ -494,33 +497,36 @@ class Instrument:
         if type(filename) is not list:
             filename = [filename]
         for f in filename:
-            if isdir(f):
+            if os.path.isdir(f):
                 self.read_calibration_dir(f, immersed)
-            else:
-                _, ext = splitext(f)
+            elif os.path.isfile(f):
+                ext = os.path.splitext(f)[1]
                 if ext in self.VALID_SIP_EXTENSIONS:
                     self.read_sip_file(f, immersed)
                 elif ext in self.VALID_CAL_EXTENSIONS:
                     self.read_calibration_file(f, immersed)
                 else:
-                    raise CalibrationFileExtensionError('File extension incorrect')
+                    raise CalibrationFileExtensionError(f'File extension incorrect: {f}')
+            else:
+                raise FileNotFoundError(f'No such file or directory: {f}')
 
     def read_calibration_file(self, filename, immersed=False):
-        _, ext = splitext(filename)
+        ext = os.path.splitext(filename)[1]
         if ext in self.VALID_CAL_EXTENSIONS:
             foo = Parser(filename, immersed)
             self.cal[foo.frame_header] = foo
             self.max_frame_header_length = max(self.max_frame_header_length, len(foo.frame_header))
         else:
-            raise CalibrationFileExtensionError('File extension incorrect')
+            raise CalibrationFileExtensionError(f'File extension incorrect: {f}')
 
     def read_calibration_dir(self, dirname, immersed=False):
         empty_dir = True
-        for fn in listdir(dirname):
-            if isfile(join(dirname, fn)) and splitext(fn)[1] in self.VALID_CAL_EXTENSIONS and basename(fn)[0] != '.':
+        for fn in os.listdir(dirname):
+            if os.path.isfile(os.path.join(dirname, fn)) and \
+                    os.path.splitext(fn)[1] in self.VALID_CAL_EXTENSIONS and os.path.basename(fn)[0] != '.':
                 # File exist, valide extension, and not hidden file
                 empty_dir = False
-                foo = Parser(join(dirname, fn), immersed)
+                foo = Parser(os.path.join(dirname, fn), immersed)
                 self.cal[foo.frame_header] = foo
                 self.max_frame_header_length = max(self.max_frame_header_length, len(foo.frame_header))
         if empty_dir:
@@ -529,13 +535,13 @@ class Instrument:
     def read_sip_file(self, filename, immersed=False):
         empty_sip = True
         archive = zipfile.ZipFile(filename, 'r')
-        dirsip = dirname(filename)
+        dirsip = os.path.dirname(filename)
         archive.extractall(path=dirsip)
         for fn in archive.namelist():
-            if splitext(fn)[1] in self.VALID_CAL_EXTENSIONS and basename(fn)[0] != '.':
+            if os.path.splitext(fn)[1] in self.VALID_CAL_EXTENSIONS and os.path.basename(fn)[0] != '.':
                 # Valide extension and not hidden file
                 empty_sip = False
-                foo = Parser(join(dirsip, fn), immersed)
+                foo = Parser(os.path.join(dirsip, fn), immersed)
                 self.cal[foo.frame_header] = foo
                 self.max_frame_header_length = max(self.max_frame_header_length, len(foo.frame_header))
         if empty_sip:
@@ -725,9 +731,9 @@ class Instrument:
             # No frame found
             return bytearray(), None, bytearray(), buffer
 
-    def parse_frame(self, frame, frame_header=None, flag_get_auxiliary_variables=None, flag_get_unusable_variables=False):
+    def parse_frame(self, frame, frame_header=None, *args, **kwargs):
         if not frame_header:
-            # Attempt to guess frame header from Standard 10 caracters SAT headers
+            # Attempt to guess frame header from Standard 10 characters SAT headers
             frame_header = frame[0:10].decode(self.ENCODING, self.UNICODE_HANDLING)
             if frame_header not in self.cal.keys():
                 # Attempt to find one of the know frame header
@@ -740,46 +746,56 @@ class Instrument:
                 if not frame_header:
                     raise FrameHeaderNotFoundError('Frame header not found or parser missing.')
         parser = self.cal[frame_header]
+        raw_decoded = self.decode(frame, parser)
+        return self.calibrate(raw_decoded, parser, frame, *args, **kwargs)
+
+    @staticmethod
+    def decode(frame, parser):
         if parser.variable_frame_length:
             # Variable length frame
             rd = list()
             # Decode value of each field
-            frame = frame[parser.frame_header_length+1:].decode(self.ENCODING, self.UNICODE_HANDLING)  # skip first value separator (comma)
+            frame = frame[parser.frame_header_length + 1:].decode(Instrument.ENCODING,
+                                                                  Instrument.UNICODE_HANDLING)  # skip first value separator (comma)
             for s, t in zip(parser.field_separator[1:], parser.data_type[:-1]):
                 index_sep = frame.find(s)
                 if index_sep == -1:
                     valid_frame = False
                     continue
-                rd.append(self._decode_ascii_data(frame[0:index_sep], t, force_ascii=True))
+                rd.append(Instrument._decode_ascii_data(frame[0:index_sep], t, force_ascii=True))
                 frame = frame[index_sep + 1:]
         else:
             # Fixed length frame
             # Decode binary data
             if parser.frame_length != len(frame):
                 raise FrameLengthError('Unexpected frame length: %s expected %d actual %d' %
-                                       (frame_header, parser.frame_length, len(frame)))
+                                       (parser.frame_header, parser.frame_length, len(frame)))
             rd = unpack(parser.frame_fmt, frame[10:])
             # Decode ASCII variables
-            rd = [self._decode_ascii_data(v, t) for v, t in zip(rd, parser.data_type)]
+            rd = [Instrument._decode_ascii_data(v, t) for v, t in zip(rd, parser.data_type)]
+        return rd
+
+    @staticmethod
+    def calibrate(rd, parser, frame=None, flag_get_auxiliary_variables=None, flag_get_unusable_variables=False):
         # Get integration time if available as required from OPTIC3 fit
         if 'INTTIME' in parser.type:
             i = parser.type.index('INTTIME')
-            aint = self._fit_data(rd[i], parser.fit_type[i], parser.cal_coefs[i], immersed=parser.immersed)
+            aint = Instrument._fit_data(rd[i], parser.fit_type[i], parser.cal_coefs[i], immersed=parser.immersed)
         else:
             aint = None
         # Core Variables (same data and fit types, serialize process in numpy)
         if parser.core_variables:
-            d = {parser.core_groupname: self._fit_data(np.array(itemgetter(*parser.core_variables)(rd)),
-                                                       parser.fit_type[parser.core_variables[0]],
-                                                       parser.core_cal_coefs, aint, parser.immersed)}
+            d = {parser.core_groupname: Instrument._fit_data(np.array(itemgetter(*parser.core_variables)(rd)),
+                                                             parser.fit_type[parser.core_variables[0]],
+                                                             parser.core_cal_coefs, aint, parser.immersed)}
         else:
             d = dict()
         # Unusable Variables (same data and fit types, serialize process in numpy)
         if flag_get_unusable_variables and parser.unusable_variables:
-            d[parser.unusable_groupname] = self._fit_data(np.array(itemgetter(*parser.unusable_variables)(rd)),
-                                                          parser.fit_type[parser.unusable_variables[0]],
-                                                          parser.unusable_cal_coefs,
-                                                          aint, parser.immersed)
+            d[parser.unusable_groupname] = Instrument._fit_data(np.array(itemgetter(*parser.unusable_variables)(rd)),
+                                                                parser.fit_type[parser.unusable_variables[0]],
+                                                                parser.unusable_cal_coefs,
+                                                                aint, parser.immersed)
         # Auxiliary variables (default: off: if core_variables | on: if no core variables)
         if (flag_get_auxiliary_variables is None and not parser.core_variables) or flag_get_auxiliary_variables:
             valid_frame = True
@@ -800,20 +816,21 @@ class Instrument:
                         # if rd[j] != self.compute_nmea_check_sum(frame):
                         #     valid_frame = False
                         pass
-                    else:
-                        if rd[j] != self.compute_check_sum(frame, parser.check_sum_index):
+                    elif frame is not None:
+                        if rd[j] != Instrument.compute_check_sum(frame, parser.check_sum_index):
                             valid_frame = False
-                d[parser.key[j]] = self._fit_data(rd[j], parser.fit_type[j], parser.cal_coefs[j], aint,
-                                                  parser.immersed)
+                d[parser.key[j]] = Instrument._fit_data(rd[j], parser.fit_type[j], parser.cal_coefs[j], aint,
+                                                        parser.immersed)
         else:
             valid_frame = None
         return d, valid_frame
 
-    def _decode_ascii_data(self, value, data_type, force_ascii=False):
+    @staticmethod
+    def _decode_ascii_data(value, data_type, force_ascii=False):
         if data_type in ['AS', 'AI', 'AI16', 'AF']:
             try:
                 # Convert from byte to string
-                foo = value.decode(self.ENCODING, self.UNICODE_HANDLING)
+                foo = value.decode(Instrument.ENCODING, Instrument.UNICODE_HANDLING)
             except (UnicodeDecodeError, AttributeError):
                 foo = value
             if data_type == 'AI':
@@ -950,6 +967,85 @@ class Instrument:
             checksum ^= ord(s)
         return checksum
 
+    def read_satview(self, filename, **kwargs):
+        """
+        Read SatView raw file
+        Assume frame is valid if parse_frame return valid=None
+
+        :param filename: SatView file to read
+        :param kwargs: key word arguments for parse_frame
+        :param debug: display unknown bytes and unknown frames
+        :return: list of FrameContainers, metadata dictionary
+        """
+        data = []
+        ignored_bytes, invalid_timestamps = 0, 0
+        valid_frames, invalid_frames = dict(), dict()
+        sathdr_present = False
+        with open(filename, 'rb') as f:
+            buffer = bytearray(f.read(32768))
+            while buffer:
+                frame = True
+                while frame or unknown_bytes:
+                    # Get Frame
+                    frame, frame_header, buffer, unknown_bytes = self.find_frame(buffer)
+                    if unknown_bytes:
+                        if unknown_bytes[:6] == b'SATHDR':
+                            sathdr_present = True
+                        else:
+                            # print(unknown_bytes)
+                            ignored_bytes += len(unknown_bytes)
+                            if frame_header and not frame:
+                                if frame_header not in invalid_frames.keys():
+                                    invalid_frames[frame_header] = 1
+                                else:
+                                    invalid_frames[frame_header] += 1
+                    if frame:
+                        if len(buffer) >= 7:
+                            # Get SatView timestamp
+                            ts = buffer[:7]
+                            buffer = buffer[7:]
+                            # Parse timestamp
+                            ts = unpack('!ii', b'\x00' + ts)
+                            try:
+                                timestamp = datetime.strptime(f'{ts[0]}{ts[1]:09d}000', '%Y%j%H%M%S%f')
+                            except ValueError as e:
+                                warnings.warn('Time Impossible, frame likely corrupted.')
+                                timestamp = 'NaN'
+                                invalid_timestamps += 1
+                            # Calibrate frame
+                            parsed_frame, valid = self.parse_frame(frame, frame_header, **kwargs)
+                            if valid == True or valid is None:
+                                data.append(FrameContainer(header=frame_header, data=parsed_frame,
+                                                           timestamp=timestamp, valid=valid))
+                                if frame_header not in valid_frames.keys():
+                                    valid_frames[frame_header] = 1
+                                else:
+                                    valid_frames[frame_header] += 1
+                            else:
+                                if frame_header not in invalid_frames.keys():
+                                    # print(frame)
+                                    invalid_frames[frame_header] = 1
+                                else:
+                                    invalid_frames[frame_header] += 1
+                        else:
+                            # Missing data to read timestamp from SatView
+                            # Refill buffer and go grab more data
+                            buffer = frame + buffer
+                            break
+                tmp_buffer = f.read(32768)
+                if tmp_buffer:
+                    buffer.extend(tmp_buffer)
+                else:
+                    break
+        # Read metadata from file
+        stat = os.stat(filename)
+        meta = dict(file_name=os.path.basename(filename), file_size=stat.st_size,
+                    file_created=datetime.utcfromtimestamp(stat.st_ctime).replace(tzinfo=timezone.utc),
+                    file_modified=datetime.utcfromtimestamp(stat.st_mtime).replace(tzinfo=timezone.utc),
+                    sathdr_present=sathdr_present, valid_frames=valid_frames, invalid_frames=invalid_frames,
+                    invalid_timestamps=invalid_timestamps, ignored_bytes=ignored_bytes)
+        return data, meta
+
     def __str__(self):
         foo = ""
         for c in self.cal.values():
@@ -996,7 +1092,7 @@ class SatViewRawToCSV:
     def run(self, filename):
         with open(filename, 'rb') as f:
             # Open output csv files
-            [filename, _] = splitext(filename)
+            [filename, _] = os.path.splitext(filename)
             for k, cal in self.instrument.cal.items():
                 self.w[k] = CSVWriter()
                 disp_key = cal.key.copy()
@@ -1018,6 +1114,9 @@ class SatViewRawToCSV:
             while data:
                 self.data_read(data)
                 data = f.read(self.READ_SIZE)
+            # Close files
+            for k in self.instrument.cal.keys():
+                self.w[k].close()
 
     # Method using check frame (less elegant)
     # def data_read(self, data):
@@ -1086,7 +1185,8 @@ class SatViewRawToCSV:
         try:
             if not frame_header:
                 frame_header = frame[0:10].decode(self.instrument.ENCODING, self.instrument.UNICODE_HANDLING)
-            [parsed_frame, valid_frame] = self.instrument.parse_frame(frame, frame_header, True)
+            [parsed_frame, valid_frame] = self.instrument.parse_frame(frame, frame_header,
+                                                                      flag_get_auxiliary_variables=True)
         except FrameHeaderNotFoundError:
             if frame_header not in self.missing_frame_header:
                 self.missing_frame_header.append(frame_header)
